open Format
open Exp

(* Evaluation by cheating *)
(* An evaluation context is a map from strings to semantic values. *)

type sem = 
     Int of int
  |  Fun of (sem -> sem)

let apply = function
     Fun f -> f
  | _ -> Support.Error.bug "apply"

let mk_int = function
    Int c -> c
  | _ -> Support.Error.bug "mk_int"

let print = function
     Int c -> print_int c
  |  Fun _ -> print_string "<fun>"

(* The linearity business isn't used at the moment *)

type context = (sem * Typ.aspect) Support.StringMap.t

let empty_context = Support.StringMap.empty

let bind x v a ctx =  Support.StringMap.add x (v,a) 
                             (Support.StringMap.remove x ctx)

let lookup x ctx =  Support.StringMap.find x ctx 

let rec choose_fresh_var x ctx =
  try 
    let b = Support.StringMap.find x ctx in
    match b with
      (_,_) -> choose_fresh_var (x ^ "'") ctx
  with
    Not_found -> x

let rename_binding ctx (x,e) =
  let x' = choose_fresh_var x ctx in
  if x = x' 
    then (x, e)
    else (x', subst1 x (Var(Support.Error.unknown,x')) e)

(* ------------------------------ *)


let recursor = Fun(function n -> 
               Fun(function g ->
               Fun(function h ->
              let rec f n = 
                 if n = 0
                    then g         
                 else  (apply (apply h (Int n)) (f (n/2)))
              in f (mk_int n))))


let eval ctx e =
  let rec ev ctx e = match e with
    Var(fi,x) -> 
      (try (match lookup x ctx with
            (v,_) -> v) with
           Not_found -> Support.Error.errAt fi 
                           ("Eval:variable " ^ x ^ " is unbound"))
  | Con(fi,c) -> 
        Int(c)
  | SZero(fi) -> Fun(function v -> Int(2*(mk_int v)))
  | SOne(fi) -> Fun(function v -> Int(2*(mk_int v)+1))
  | Abs(fi,x,a,ty,e1) -> 
          let (x,e1) = rename_binding ctx (x,e1) in
             Fun(function v -> ev (bind x v a ctx) e1)
  | Let(fi,x,a,e1,e2) -> 
      let v1 = ev ctx e1 in
      let (x,e2) = rename_binding ctx (x,e2) in
      ev (bind x  v1 a ctx) e2 
  | App(fi,e1,e2) ->
      apply (ev ctx e1) (ev ctx e2)
  | Case(fi,ty,e1,e2,e3,e4) -> 
       let c = mk_int (ev ctx e1) in
              if c = 0 
                  then ev ctx e2
              else if c mod 2 = 0 
                  then apply (ev ctx e3) (Int (c/2))
              else apply (ev ctx e4) (Int (c/2))
  | Saferec(fi,ty) -> recursor
  | Linrec(fi,ty) -> recursor
  | Coerce(fi,e,_)  -> ev ctx e
      
in ev ctx e

 





